from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, overload

from fieldz._types import DataclassParams, Field

if TYPE_CHECKING:
    import msgspec
    from typing_extensions import TypeGuard


@overload
def is_msgspec_struct(obj: type) -> TypeGuard[type[msgspec.Struct]]:
    ...


@overload
def is_msgspec_struct(obj: object) -> TypeGuard[msgspec.Struct]:
    ...


def is_msgspec_struct(obj: Any) -> bool:
    """Return True if the class is a `msgspec.Struct`."""
    msgspec = sys.modules.get("msgspec", None)
    cls = obj if isinstance(obj, type) else type(obj)
    return msgspec is not None and issubclass(cls, msgspec.Struct)


is_instance = is_msgspec_struct


def asdict(obj: msgspec.Struct) -> dict[str, Any]:
    import msgspec.structs

    return msgspec.structs.asdict(obj)


def astuple(obj: msgspec.Struct) -> tuple[Any, ...]:
    import msgspec.structs

    return msgspec.structs.astuple(obj)


def replace(obj: msgspec.Struct, /, **changes: Any) -> Any:
    """Return a copy of obj with the specified changes."""
    import msgspec.structs

    return msgspec.structs.replace(obj, **changes)


def fields(obj: msgspec.Struct | type[msgspec.Struct]) -> tuple:
    import msgspec

    return tuple(
        Field(
            name=f.name,
            type=f.type,
            default=(Field.MISSING if f.default is msgspec.NODEFAULT else f.default),
            default_factory=(
                Field.MISSING
                if f.default_factory is msgspec.NODEFAULT
                else f.default_factory
            ),
            native_field=f,
        )
        for f in msgspec.structs.fields(obj)
    )


def params(obj: msgspec.Struct) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    cfg: msgspec.structs.StructConfig | None = getattr(obj, "__struct_config__", None)
    if cfg is None:
        return DataclassParams()  # pragma: no cover

    # this will be covered in msgspec > 0.13.1
    return DataclassParams(
        frozen=cfg.frozen, eq=cfg.eq, order=cfg.order, init=True, repr=True
    )



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, ClassVar, Protocol

from fieldz._types import DataclassParams, Field

if TYPE_CHECKING:

    class _DataclassParams(Protocol):
        init: bool
        repr: bool
        eq: bool
        order: bool
        unsafe_hash: bool
        frozen: bool

    class DataclassInstance(Protocol):
        __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]
        __dataclass_params__: _DataclassParams


def is_instance(obj: Any) -> bool:
    return bool(dataclasses.is_dataclass(obj))


def asdict(obj: Any) -> dict[str, Any]:
    return dataclasses.asdict(obj)


def astuple(obj: Any) -> tuple[Any, ...]:
    return dataclasses.astuple(obj)


def replace(obj: Any, /, **changes: Any) -> Any:
    """Return a copy of obj with the specified changes."""
    return dataclasses.replace(obj, **changes)


def fields(obj: Any | type[Any]) -> tuple[Field, ...]:
    """Return a tuple of fields for the class or instance."""
    return tuple(
        Field(
            name=f.name,
            type=f.type,
            default=(
                f.default if f.default is not dataclasses.MISSING else Field.MISSING
            ),
            default_factory=(
                f.default_factory if callable(f.default_factory) else Field.MISSING
            ),
            init=f.init,
            repr=f.repr,
            hash=f.hash,
            compare=f.compare,
            metadata=f.metadata,
            native_field=f,
        )
        for f in dataclasses.fields(obj)
    )


def params(obj: Any) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    params: _DataclassParams | None = getattr(obj, "__dataclass_params__", None)
    if params is not None:
        return DataclassParams(
            init=params.init,
            repr=params.repr,
            eq=params.eq,
            order=params.order,
            unsafe_hash=params.unsafe_hash,
            frozen=params.frozen,
        )
    return DataclassParams()  # pragma: no cover



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from . import adapters

if TYPE_CHECKING:
    from ._types import DataclassParams, Field


def asdict(obj: Any) -> dict[str, Any]:
    """Return a dict representation of obj."""
    return get_adapter(obj).asdict(obj)


def astuple(obj: Any) -> tuple[Any, ...]:
    """Return a tuple representation of obj."""
    return get_adapter(obj).astuple(obj)


def replace(obj: Any, /, **changes: Any) -> Any:
    """Return a copy of obj with the specified changes."""
    return get_adapter(obj).replace(obj, **changes)


def fields(obj: Any | type[Any], *, parse_annotated: bool = True) -> tuple[Field, ...]:
    """Return a tuple of fields for the class or instance."""
    fields = get_adapter(obj).fields(obj)
    if parse_annotated:
        fields = tuple(field.parse_annotated() for field in fields)
    return fields


def params(obj: Any) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    return get_adapter(obj).params(obj)


# Order matters here. The first adapter to return True for is_instance will be used.
ADAPTERS: tuple[adapters.Adapter, ...] = (
    adapters._pydantic,
    adapters._attrs,
    adapters._msgspec,
    adapters._dataclassy,
    adapters._dataclasses,
    adapters._named_tuple,
    adapters._typed_dict,
)


def get_adapter(obj: Any) -> adapters.Adapter:
    """Return the module of the given object."""
    for mod in ADAPTERS:
        if mod.is_instance(obj):
            return mod
    raise TypeError(f"Unsupported dataclass type: {type(obj)}")  # pragma: no cover



from __future__ import annotations

import dataclasses
import re
import sys
from typing import TYPE_CHECKING, Any, Iterator, overload

from fieldz._types import (
    Constraints,
    DataclassParams,
    Field,
    _is_annotated_type,
    _parse_annotatedtypes_meta,
)

if TYPE_CHECKING:
    import pydantic
    import pydantic.fields
    from typing_extensions import TypeGuard


@overload
def is_pydantic_model(obj: type) -> TypeGuard[type[pydantic.BaseModel]]:
    ...


@overload
def is_pydantic_model(obj: object) -> TypeGuard[pydantic.BaseModel]:
    ...


def is_pydantic_model(obj: Any) -> bool:
    """Return True if obj is a pydantic.BaseModel subclass or instance."""
    pydantic = sys.modules.get("pydantic", None)
    cls = obj if isinstance(obj, type) else type(obj)
    if pydantic is not None and issubclass(cls, pydantic.BaseModel):
        return True
    elif hasattr(cls, "__pydantic_model__") or hasattr(cls, "__pydantic_fields__"):
        return True
    return False


is_instance = is_pydantic_model


def asdict(obj: pydantic.BaseModel) -> dict[str, Any]:
    # sourcery skip: reintroduce-else

    if hasattr(obj, "__dataclass_params__"):
        return dataclasses.asdict(obj)

    if hasattr(obj, "model_dump"):
        return obj.model_dump()
    return obj.dict()


def astuple(obj: pydantic.BaseModel) -> tuple[Any, ...]:
    return tuple(asdict(obj).values())


def replace(obj: pydantic.BaseModel, /, **changes: Any) -> Any:
    """Return a copy of obj with the specified changes."""
    if hasattr(obj, "__dataclass_params__"):
        return dataclasses.replace(obj, **changes)  # type: ignore

    if hasattr(obj, "model_copy"):
        return obj.model_copy(update=changes)
    return obj.copy(update=changes)


def _fields_v1(obj: pydantic.BaseModel | type[pydantic.BaseModel]) -> Iterator[Field]:
    from pydantic.fields import Undefined  # type: ignore

    annotations = getattr(obj, "__annotations__", {})
    for name, modelfield in obj.__fields__.items():  # type: ignore
        factory = (
            modelfield.default_factory
            if callable(modelfield.default_factory)
            else Field.MISSING
        )
        default = (
            Field.MISSING
            if factory is not Field.MISSING
            or modelfield.default in (Undefined, Ellipsis)
            else modelfield.default
        )
        # backport from pydantic2
        _extra_dict = modelfield.field_info.extra.copy()  # type: ignore
        if "json_schema_extra" in _extra_dict:
            _extra_dict.update(_extra_dict.pop("json_schema_extra"))

        yield Field(
            name=name,
            type=annotations.get(name),  # rather than outer_type_
            default=default,
            default_factory=(factory if callable(factory) else Field.MISSING),
            native_field=modelfield,
            description=modelfield.field_info.description,  # type: ignore
            metadata=_extra_dict,
            constraints=_constraints_v1(modelfield),
        )


def _constraints_v1(modelfield: Any) -> Constraints | None:
    kwargs = {}
    # check if the type is a pydantic constrained type
    for subt in modelfield.type_.__mro__:
        if (subt.__module__ or "").startswith("pydantic.types"):
            keys = (
                "gt",
                "ge",
                "lt",
                "le",
                "multiple_of",
                "max_digits",
                "decimal_places",
                "min_length",
                "max_length",
            )
            kwargs.update({key: getattr(modelfield.type_, key, None) for key in keys})
            if regex := getattr(modelfield.type_, "regex", None):
                if isinstance(regex, re.Pattern):
                    regex = regex.pattern
                kwargs["pattern"] = regex
    return Constraints(**kwargs) if kwargs else None


def _fields_v2(obj: pydantic.BaseModel | type[pydantic.BaseModel]) -> Iterator[Field]:
    from pydantic_core import PydanticUndefined

    if hasattr(obj, "__pydantic_fields__"):  # v2 dataclass
        _fields = obj.__pydantic_fields__.items()
    else:
        _fields = obj.model_fields.items()

    annotations = getattr(obj, "__annotations__", {})
    for name, finfo in _fields:
        factory = (
            finfo.default_factory if callable(finfo.default_factory) else Field.MISSING
        )
        default = (
            Field.MISSING
            if finfo.default in (PydanticUndefined, Ellipsis)
            else finfo.default
        )
        extra = finfo.json_schema_extra

        annotated_type = annotations.get(name)
        if not _is_annotated_type(annotated_type):
            annotated_type = None

        c = _parse_annotatedtypes_meta(finfo.metadata)
        constraints = Constraints(**c) if c else None

        yield Field(
            name=name,
            type=finfo.annotation,
            default=default,
            default_factory=factory,
            native_field=finfo,
            description=finfo.description,
            metadata=extra if isinstance(extra, dict) else {},
            annotated_type=annotated_type,
            constraints=constraints,
        )


def fields(obj: pydantic.BaseModel | type[pydantic.BaseModel]) -> tuple[Field, ...]:
    if hasattr(obj, "model_fields") or hasattr(obj, "__pydantic_fields__"):
        return tuple(_fields_v2(obj))
    if hasattr(obj, "__pydantic_model__"):
        obj = obj.__pydantic_model__  # v1 dataclass
    return tuple(_fields_v1(obj))


def params(obj: pydantic.BaseModel) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    if hasattr(obj, "__dataclass_params__"):
        p = obj.__dataclass_params__
        return DataclassParams(
            frozen=p.frozen,
            init=p.init,
            repr=p.repr,
            eq=p.eq,
            order=p.order,
            unsafe_hash=p.unsafe_hash,
        )
    if hasattr(obj, "model_config"):
        cfg_dict: pydantic.ConfigDict = obj.model_config
        return DataclassParams(
            # unsafe_hash=not cfg.get("frozen", False),
            frozen=cfg_dict.get("frozen", False)
        )
    else:
        cfg = obj.__config__  # type: ignore
        return DataclassParams(frozen=cfg.allow_mutation is False)
    return DataclassParams()  # pragma: no cover



import dataclasses
from typing import Callable, List, NamedTuple, Optional, TypedDict

import pytest
from fieldz import Field, asdict, astuple, fields, params, replace
from fieldz.adapters._named_tuple import is_named_tuple


def _dataclass_model() -> type:
    @dataclasses.dataclass
    class Model:
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = dataclasses.field(default_factory=list)

    return Model


def _named_tuple() -> type:
    class Model(NamedTuple):
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = []  # noqa

    return Model


def _pydantic_model() -> type:
    from pydantic import BaseModel, Field

    class Model(BaseModel):
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = Field(default_factory=list)

    return Model


def _pydantic_dataclass() -> type:
    from pydantic.dataclasses import dataclass

    @dataclass
    class Model:
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = dataclasses.field(default_factory=list)

    return Model


def _sqlmodel() -> type:
    pytest.importorskip("sqlmodel")
    from sqlmodel import Field, SQLModel

    class Model(SQLModel):
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = Field(default_factory=list)

    return Model


def _attrs_model() -> type:
    import attr

    @attr.define
    class Model:
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = attr.field(default=attr.Factory(list))

    return Model


def _msgspec_model() -> type:
    import msgspec

    class Model(msgspec.Struct):
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = msgspec.field(default_factory=list)

    return Model


def _dataclassy_model() -> type:
    import dataclassy

    @dataclassy.dataclass
    class Model:
        a: int = 0
        b: Optional[str] = None
        c: float = 0.0
        d: bool = False
        e: List[int] = []  # noqa

    return Model


def _django_model() -> type:
    from django.db import models

    class Model(models.Model):
        a: int = models.IntegerField(default=0)
        b: str = models.CharField(default="b", max_length=255)
        c: float = models.FloatField(default=0.0)
        d: bool = models.BooleanField(default=False)
        e: List[int] = models.JSONField(default=list)

    return Model


@pytest.mark.parametrize(
    "builder",
    [
        _dataclass_model,
        _named_tuple,
        _dataclassy_model,
        _pydantic_model,
        _attrs_model,
        _msgspec_model,
        _sqlmodel,
        _pydantic_dataclass,
        # _django_model,
    ],
)
def test_adapters(builder: Callable) -> None:
    model = builder()
    obj = model()
    assert asdict(obj) == {"a": 0, "b": None, "c": 0.0, "d": False, "e": []}
    assert astuple(obj) == (0, None, 0.0, False, [])
    fields_ = fields(obj)
    assert [f.name for f in fields_] == ["a", "b", "c", "d", "e"]
    assert [f.type for f in fields_] == [int, Optional[str], float, bool, List[int]]
    assert [f.frozen for f in fields_] == [False] * 5
    if is_named_tuple(obj):
        assert [f.default for f in fields_] == [0, None, 0.0, False, []]
    else:
        # namedtuples don't have default_factory
        assert [f.default for f in fields_] == [
            0,
            None,
            0.0,
            False,
            Field.MISSING,
        ]
        assert [f.default_factory for f in fields_] == [
            *[Field.MISSING] * 4,
            list,
        ]

    obj2 = replace(obj, a=1, b="b2", c=1.0, d=True, e=[1, 2, 3])
    assert asdict(obj2) == {"a": 1, "b": "b2", "c": 1.0, "d": True, "e": [1, 2, 3]}

    p = params(obj)
    assert p.eq is True
    assert p.order is False
    assert p.repr is True
    assert p.init is True
    assert p.unsafe_hash is False
    assert p.frozen is False


def test_typed_dict() -> None:
    class Model(TypedDict):
        a: int
        b: Optional[str]
        c: float
        d: bool
        e: List[int]

    assert fields(Model) == (
        Field(name="a", type=int),
        Field(name="b", type=Optional[str]),
        Field(name="c", type=float),
        Field(name="d", type=bool),
        Field(name="e", type=List[int]),
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from fieldz._types import DataclassParams, Field


#  MISSING, replace
_T = TypeVar("_T")


DataclassLike: TypeAlias = Any
_DataclassType = TypeVar("_DataclassType", bound=DataclassLike)


class Adapter(Protocol, Generic[_DataclassType]):
    """Protocol that adapter modules need to implement."""

    # @overload
    # def is_instance(self, obj: _DataclassType) -> Literal[True]:
    #     ...

    # @overload
    # def is_instance(self, obj: type) -> TypeGuard[type[_DataclassType]]:
    #     ...

    # @overload
    # def is_instance(
    #     self, obj: object
    # ) -> TypeGuard[_DataclassType | type[_DataclassType]]:
    #     ...

    def is_instance(self, obj: DataclassLike) -> bool:
        """Return true if obj is a a recognized instance for this adapter."""

    def asdict(self, obj: DataclassLike) -> dict[str, Any]:
        """Return a dict representation of obj."""

    def astuple(self, obj: DataclassLike) -> tuple[Any, ...]:
        """Return a tuple representation of obj."""

    def replace(self, obj: _DataclassType, /, **changes: Any) -> _DataclassType:
        """Return a copy of obj with the specified changes."""

    def fields(self, obj: DataclassLike | type[DataclassLike]) -> tuple[Field, ...]:
        """Return a tuple of fields for the class or instance."""

    def params(self, obj: DataclassLike) -> DataclassParams:
        """Return parameters used to define the dataclass."""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from fieldz._types import DataclassParams, Field

if TYPE_CHECKING:
    from typing import Iterable

    from typing_extensions import Self, TypeGuard

    class NamedTupleInstance(tuple[Any, ...]):
        _field_defaults: ClassVar[dict[str, Any]]
        _fields: ClassVar[tuple[str, ...]]

        @classmethod
        def _make(cls, iterable: Iterable[Any]) -> Self:
            ...

        def _asdict(self) -> dict[str, Any]:
            ...

        def _replace(self, **kwargs: Any) -> Self:
            ...


def is_named_tuple(obj: Any) -> TypeGuard[NamedTupleInstance]:
    cls = obj if isinstance(obj, type) else type(obj)
    return (
        hasattr(cls, "_fields")
        and hasattr(cls, "_field_defaults")
        and hasattr(cls, "_make")
    )


is_instance = is_named_tuple


def asdict(obj: NamedTupleInstance) -> dict[str, Any]:
    return obj._asdict()


def astuple(obj: NamedTupleInstance) -> tuple[Any, ...]:
    return obj


def replace(obj: NamedTupleInstance, /, **changes: Any) -> NamedTupleInstance:
    """Return a copy of obj with the specified changes."""
    return obj._replace(**changes)


def fields(obj: NamedTupleInstance | type[NamedTupleInstance]) -> tuple[Field, ...]:
    """Return a tuple of fields for the class or instance."""
    annotations = getattr(obj, "__annotations__", {})
    defaults = getattr(obj, "_field_defaults", {})
    return tuple(
        Field(name=name, type=annotations.get(name, Any), default=defaults.get(name))
        for name in obj._fields
    )


def params(obj: Any) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    return DataclassParams()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from fieldz._types import DataclassParams, Field, _is_classvar

if TYPE_CHECKING:
    from typing_extensions import TypeGuard


def is_typed_dict(obj: Any) -> TypeGuard[dict]:
    # NOTE: this will ONLY work on TypedDict subclasses, not instances
    # because at runtime, TypedDict instances are simply dicts, and there's no way
    # to tell if a dict is an instance of a TypedDict subclass

    # in python 3.8 the only difference between a TypedDict subclass and a dict is:
    # - __dict__
    # - __weakref__
    # - __module__
    # - __total__
    # - __annotations_
    # ... here we choose to check for __total__ and __annotations__
    # (since most things will have __dict__, __module__, __weakref__)
    # later versions also have __required_keys__ and __optional_keys__
    return (
        isinstance(obj, type)
        and issubclass(obj, dict)
        and hasattr(obj, "__annotations__")
        and hasattr(obj, "__total__")
    )


is_instance = is_typed_dict


def asdict(obj: dict) -> dict[str, Any]:
    return obj


def astuple(obj: dict) -> tuple[Any, ...]:
    return tuple(obj.values())


def replace(obj: dict, /, **changes: Any) -> dict:
    """Return a copy of obj with the specified changes."""
    return {**obj, **changes}


def fields(obj: dict | type[dict]) -> tuple[Field, ...]:
    """Return a tuple of fields for the class or instance."""
    return tuple(
        Field(name=name, type=hint)
        for name, hint in getattr(obj, "__annotations__", {}).items()
        if not _is_classvar(hint)
    )


def params(obj: dict) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    return DataclassParams()



from __future__ import annotations

import sys
import types
import typing
from typing import Any

import typing_extensions

try:
    from typing import _TypingBase  # type: ignore[attr-defined]
except ImportError:
    from typing import _Final as _TypingBase  # type: ignore[attr-defined]

typing_base = _TypingBase

if sys.version_info < (3, 9):
    # python < 3.9 does not have GenericAlias (list[int], tuple[str, ...] and so on)
    TypingGenericAlias = ()
else:
    from typing import GenericAlias as TypingGenericAlias  # type: ignore

if sys.version_info < (3, 10):

    def origin_is_union(tp: type[Any] | None) -> bool:
        return tp is typing.Union

    WithArgsTypes = (TypingGenericAlias,)

else:

    def origin_is_union(tp: type[Any] | None) -> bool:
        return tp is typing.Union or tp is types.UnionType  # type: ignore

    WithArgsTypes = (typing._GenericAlias, types.GenericAlias, types.UnionType)  # type: ignore[attr-defined]  # noqa: E501


class PlainRepr(str):
    """String class where repr doesn't include quotes.

    Useful with Representation when you want to return a string
    representation of something that is valid (or pseudo-valid) python.
    """

    def __repr__(self) -> str:
        return str(self)

    @classmethod
    def for_type(cls, tp: Any) -> PlainRepr:
        return PlainRepr(display_as_type(tp))


def display_as_type(obj: Any, modern_union: bool = False) -> str:
    """Pretty representation of a type.

    Should be as close as possible to the original type definition string.
    Takes some logic from `typing._type_repr`.
    """
    if isinstance(obj, types.FunctionType):
        return obj.__name__
    elif obj is ...:
        return "..."
    elif obj in (None, type(None)):
        return "None"

    if not isinstance(obj, (typing_base, WithArgsTypes, type)):
        obj = obj.__class__

    if origin_is_union(typing_extensions.get_origin(obj)):
        args = [display_as_type(x) for x in typing_extensions.get_args(obj)]
        if modern_union:
            return " | ".join(args)
        if len(args) == 2 and "None" in args:
            args.remove("None")
            return f"Optional[{args[0]}]"
        return f"Union[{', '.join(args)}]"
    elif isinstance(obj, WithArgsTypes):
        argstr = ", ".join(map(display_as_type, typing_extensions.get_args(obj)))
        return f"{obj.__qualname__}[{argstr}]"
    elif isinstance(obj, type):
        return obj.__qualname__
    else:
        return repr(obj).replace("typing.", "").replace("typing_extensions.", "")



import annotated_types as at
from fieldz import fields
from pydantic import BaseModel, Field, conint, constr
from typing_extensions import Annotated

PATTERN = r"^[a-z]+$"
try:
    e_constr = constr(regex=r"^[a-z]+$")
    f_field = Field(default="abc", regex=PATTERN)
except TypeError:  # pydantic v2
    e_constr = constr(pattern=r"^[a-z]+$")  # type: ignore
    f_field = Field(default="abc", pattern=PATTERN)


def test_pydantic_constraints() -> None:
    class M(BaseModel):
        a: int = Field(default=50, ge=42, le=100)
        b: Annotated[int, Field(ge=42, le=100)] = 50
        c: Annotated[int, at.Ge(42), at.Le(100)] = 50
        d: conint(ge=42, le=100) = 50  # type: ignore
        e: e_constr = "abc"  # type: ignore
        f: str = f_field

    for f in fields(M):
        assert f.constraints
        if f.name in {"e", "f"}:
            assert f.constraints.pattern == PATTERN
        else:
            assert f.constraints.ge == 42
            assert f.constraints.le == 100
            assert f.default == 50
            assert f.default == 50



"""Adapter modules for fieldz."""
from . import (
    _attrs,
    _dataclasses,
    _dataclassy,
    _msgspec,
    _named_tuple,
    _pydantic,
    _typed_dict,
)
from .protocol import Adapter

__all__ = [
    "Adapter",
    "_attrs",
    "_dataclasses",
    "_dataclassy",
    "_msgspec",
    "_named_tuple",
    "_pydantic",
    "_typed_dict",
]



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, ClassVar, Protocol

from fieldz._types import DataclassParams, Field

if TYPE_CHECKING:
    from typing_extensions import TypedDict, TypeGuard

    class DataclassyParams(TypedDict):
        init: bool
        repr: bool
        eq: bool
        order: bool
        unsafe_hash: bool
        frozen: bool
        kw_only: bool
        match_args: bool
        hide_internals: bool
        iter: bool
        kwargs: bool
        slots: bool

    class DataclassyInstance(Protocol):
        __dataclass__: ClassVar[DataclassyParams]


def is_instance(obj: Any) -> TypeGuard[DataclassyInstance | type[DataclassyInstance]]:
    if TYPE_CHECKING:
        from dataclassy import functions
    else:
        functions = sys.modules.get("dataclassy.functions", None)

    return False if functions is None else functions.is_dataclass(obj)


def asdict(obj: Any) -> dict[str, Any]:
    import dataclassy

    return dataclassy.as_dict(obj)


def astuple(obj: Any) -> tuple[Any, ...]:
    import dataclassy

    return dataclassy.as_tuple(obj)


def replace(obj: Any, /, **changes: Any) -> Any:
    """Return a copy of obj with the specified changes."""
    import dataclassy

    return dataclassy.replace(obj, **changes)


def fields(obj: Any | type) -> tuple:
    import dataclassy

    defaults = getattr(obj, "__defaults__", None) or getattr(obj, "__dict__", {})
    fields = []
    for name, type_ in dataclassy.fields(obj).items():
        default = defaults.get(name, Field.MISSING)
        default_factory: Any = Field.MISSING
        # this is just how dataclassy does it... (you can assign a mutable default)
        # but, to match the other adapters, we perform a little checking to
        # normalize common cases to a "typical" default_factory
        if hasattr(default, "copy") and callable(default.copy):
            if default == []:
                default_factory = list
            elif default == {}:
                default_factory = dict
            elif default == ():
                default_factory = tuple
            elif default == set():
                default_factory = set
            else:
                # the general fallback case
                default_factory = default.copy
            default = Field.MISSING

        fields.append(
            Field(
                name=name,
                type=type_,
                default=default,
                default_factory=default_factory,
            )
        )
    return tuple(fields)


def params(obj: Any) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    params: DataclassyParams | None = getattr(obj, "__dataclass__", None)

    if params is not None:
        return DataclassParams(
            init=params["init"],
            repr=params["repr"],
            eq=params["eq"],
            order=params["order"],
            unsafe_hash=params["unsafe_hash"],
            frozen=params["frozen"],
        )
    return DataclassParams()  # pragma: no cover



from dataclasses import dataclass
from typing import List

import annotated_types as at
from fieldz import fields
from typing_extensions import Annotated


def test_annotated_types() -> None:
    @dataclass
    class MyClass:
        age: Annotated[int, at.Gt(18)]
        span: Annotated[float, at.Interval(ge=0, le=10)]
        even: Annotated[int, at.MultipleOf(2)]
        # note that annotated types will NOT iterate min_length=0
        my_list: Annotated[List[int], at.Len(1, 10)]
        lower_name: Annotated[str, at.Predicate(str.islower)]
        # TODO: determine how to handle nested Annotated types
        # factors: List[Annotated[str, at.Predicate(str.islower)]]
        with_tz: Annotated[str, at.Timezone(...)]
        unannotated: int = 0

    fields_ = {f.name: f for f in fields(MyClass)}
    assert fields_["age"].constraints.gt == 18
    assert fields_["span"].constraints.ge == 0
    assert fields_["span"].constraints.le == 10
    assert fields_["even"].constraints.multiple_of == 2
    assert fields_["my_list"].constraints.min_length == 1
    assert fields_["my_list"].constraints.max_length == 10
    assert fields_["lower_name"].constraints.predicate == str.islower
    assert fields_["lower_name"].constraints.predicate == str.islower


def test_msgspec_constraints() -> None:
    from msgspec import Meta, Struct

    # TODO: determine how to handle nested Annotated types
    # UnixName = Annotated[
    #     str, Meta(min_length=1, max_length=32, pattern="^[a-z_][a-z0-9_-]*$")
    # ]
    # groups: Annotated[set[UnixName], Meta(max_length=16)] = set()  # type: ignore

    class MyClass(Struct):
        age: Annotated[int, Meta(gt=18)]
        span: Annotated[float, Meta(ge=0, le=10)]
        even: Annotated[int, Meta(multiple_of=2)]
        my_list: Annotated[List[int], Meta(min_length=1, max_length=10)]
        # msgspec puts title and description in the Meta object
        with_title: Annotated[str, Meta(title="Title", description="Description")]
        with_tz: Annotated[int, Meta(tz=True)]

    fields_ = {f.name: f for f in fields(MyClass)}
    assert fields_["age"].constraints.gt == 18
    assert fields_["span"].constraints.ge == 0
    assert fields_["span"].constraints.le == 10
    assert fields_["even"].constraints.multiple_of == 2
    assert fields_["my_list"].constraints.min_length == 1
    assert fields_["my_list"].constraints.max_length == 10
    assert fields_["with_title"].title == "Title"
    assert fields_["with_title"].description == "Description"
    assert fields_["with_tz"].constraints.tz is True



"""Utilities for providing compatibility with many dataclass-like libraries."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("fieldz")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

__all__ = [
    "Adapter",
    "asdict",
    "astuple",
    "DataclassParams",
    "Field",
    "fields",
    "get_adapter",
    "params",
    "replace",
]

from ._functions import asdict, astuple, fields, get_adapter, params, replace
from ._types import DataclassParams, Field
from .adapters import Adapter



from __future__ import annotations

import dataclasses
import enum
import sys
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Iterable,
    Literal,
    Mapping,
    TypeVar,
)

# python 3.8's `get_origin` is not Annotated-aware
from typing_extensions import Annotated, get_args, get_origin

from fieldz._repr import PlainRepr

if TYPE_CHECKING:
    import builtins

    from typing_extensions import Self

_T = TypeVar("_T")

DC_KWARGS = {"frozen": True}
if sys.version_info >= (3, 10):
    DC_KWARGS["slots"] = True


class _MISSING_TYPE(enum.Enum):
    MISSING = "MISSING"

    def __repr__(self) -> str:
        return self.value


@dataclasses.dataclass(**DC_KWARGS)
class DataclassParams:
    init: bool = True
    repr: bool = True
    eq: bool = True
    order: bool = False
    unsafe_hash: bool = False
    frozen: bool = False


@dataclasses.dataclass(**DC_KWARGS)
class Constraints:
    gt: int | float | None = None
    ge: int | float | None = None
    lt: int | float | None = None
    le: int | float | None = None
    multiple_of: int | float | None = None
    min_length: int | None = None  # for str
    max_length: int | None = None  # for str
    max_digits: int | None = None  # for decimal
    decimal_places: int | None = None  # for decimal
    pattern: str | None = None
    deprecated: bool | None = None
    tz: bool | None = None
    predicate: Callable[[Any], bool] | None = None
    # enum: list[Any] | None = None
    # const: Any | None = None

    def __rich_repr__(self) -> Iterable[tuple[str, Any]]:
        for name, val in dataclasses.asdict(self).items():
            if val is not None:
                yield name, val


@dataclasses.dataclass(**DC_KWARGS)
class Field(Generic[_T]):
    MISSING: ClassVar[Literal[_MISSING_TYPE.MISSING]] = _MISSING_TYPE.MISSING

    name: str
    type: type[_T] | None = None
    description: str | None = None
    title: str | None = None
    default: _T | Literal[_MISSING_TYPE.MISSING] = MISSING
    default_factory: Callable[[], _T] | Literal[_MISSING_TYPE.MISSING] = MISSING
    repr: bool = True
    hash: bool | None = None
    init: bool = True
    compare: bool = True
    metadata: Mapping[Any, Any] = dataclasses.field(default_factory=dict)
    kw_only: bool = False
    # extra
    frozen: bool = False
    native_field: Any | None = dataclasses.field(
        default=None, compare=False, repr=False
    )
    constraints: Constraints | None = None

    # populated during parse_annotated
    annotated_type: builtins.type[_T] | None = dataclasses.field(
        default=None, repr=False, compare=False
    )

    def __rich_repr__(self) -> Iterable[tuple[str, Any]]:
        for f in dataclasses.fields(self):
            if not f.repr:
                continue
            val = getattr(self, f.name)
            if f.name == "type":
                val = PlainRepr.for_type(val)
            elif f.name in {"default_factory", "default"} and val is Field.MISSING:
                continue
            yield f.name, val

    def parse_annotated(self) -> Self:
        """Extract info from Annotated type if present, and return new field.

        If `self.type` is not a `typing.Annotated` type, return self unchanged.
        """
        if not _is_annotated_type(self.type):
            return self

        kwargs, constraints = _parse_annotated_hint(self.type)

        for key in ("default", "name"):
            if (val := getattr(self, key)) not in (Field.MISSING, None) and kwargs.get(
                key
            ) not in (Field.MISSING, None):
                warnings.warn(
                    f"Cannot set {key!r} in both type annotation and field. Overriding "
                    f"{key!r}={kwargs[key]!r} with {val!r}.",
                    stacklevel=2,
                )

        if self.constraints is not None:
            kwargs["constraints"] = dataclasses.replace(self.constraints, **constraints)
        elif constraints:
            kwargs["constraints"] = Constraints(**constraints)

        return dataclasses.replace(self, **kwargs)


def _parse_annotated_hint(hint: Any) -> tuple[dict, dict]:
    """Convert an Annotated type to a dict of Field kwargs."""
    # hint should have been checked to be an Annotated[...] type
    origin, *metadata = get_args(hint)
    kwargs: dict[str, Any] = {"type": origin, "annotated_type": hint}
    constraints: dict[str, Any] = {}

    # deal with annotated_types
    constraints.update(_parse_annotatedtypes_meta(metadata))

    # deal with msgspec
    m_kwargs, m_constraints = _parse_msgspec_meta(metadata)
    kwargs.update(m_kwargs)
    constraints.update(m_constraints)

    # TODO: support pydantic.fields.FieldInfo?
    # TODO: support re.Pattern?
    # TODO: support msgspec
    return kwargs, constraints


# At the moment, all of our constraint names match msgspec.Meta attributes
# (we are a superset of msgspec.Meta)
CONSTRAINT_NAMES = {f.name for f in dataclasses.fields(Constraints)}
FIELD_NAMES = {f.name for f in dataclasses.fields(Field)}


def _parse_annotatedtypes_meta(metadata: list[Any]) -> dict[str, Any]:
    """Extract constraints from annotated_types metadata."""
    if TYPE_CHECKING:
        import annotated_types as at
    else:
        at = sys.modules.get("annotated_types")
        if at is None:
            return {}  # pragma: no cover

    a_kwargs = {}
    for item in metadata:
        # annotated_types >= 0.3.0 is supported
        if isinstance(item, (at.BaseMetadata, at.GroupedMetadata)):
            values = {k: getattr(item, k) for k in CONSTRAINT_NAMES if hasattr(item, k)}
            a_kwargs.update(values)
            # annotated types calls the value of a Predicate "func"
            if hasattr(item, "func"):
                a_kwargs["predicate"] = item.func

            # these were changed in v0.4.0
            if hasattr(item, "min_inclusive"):  # pragma: no cover
                a_kwargs["min_length"] = item.min_inclusive
            if hasattr(item, "max_exclusive"):  # pragma: no cover
                a_kwargs["max_length"] = item.max_exclusive - 1
    return a_kwargs


def _parse_msgspec_meta(metadata: list[Any]) -> tuple[dict, dict]:
    """Extract constraints from msgspec.Meta metadata."""
    if TYPE_CHECKING:
        import msgspec
    else:
        msgspec = sys.modules.get("msgspec")
        if msgspec is None:
            return {}, {}

    field_kwargs = {}
    constraints = {}
    for item in metadata:
        if isinstance(item, msgspec.Meta):
            constraints.update(
                {
                    k: val
                    for k in CONSTRAINT_NAMES
                    if (val := getattr(item, k, None)) is not None
                }
            )
            field_kwargs.update(
                {
                    k: val
                    for k in FIELD_NAMES
                    if (val := getattr(item, k, None)) is not None
                }
            )

    return field_kwargs, constraints


def _is_annotated_type(hint: Any) -> bool:
    """Whether the field is an Annotated type."""
    return get_origin(hint) is Annotated


def _is_classvar(a_type: Any) -> bool:
    return a_type is ClassVar or get_origin(a_type) is ClassVar


def _is_initvar(a_type: Any) -> bool:
    return a_type is dataclasses.InitVar or type(a_type) is dataclasses.InitVar



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, ClassVar, Protocol, cast, overload

from fieldz._types import DataclassParams, Field

if TYPE_CHECKING:
    import attrs
    from typing_extensions import TypeGuard

    class AttrsInstance(Protocol):
        __attrs_attrs__: ClassVar[tuple[attrs.Attribute, ...]]


@overload
def is_attrs_class(obj: type) -> TypeGuard[type[AttrsInstance]]:
    ...


@overload
def is_attrs_class(obj: object) -> TypeGuard[AttrsInstance]:
    ...


def is_attrs_class(obj: object) -> bool:
    """Return True if the class is an attrs class."""
    attr = sys.modules.get("attr", None)
    cls = obj if isinstance(obj, type) else type(obj)
    return attr.has(cls) if attr is not None else False  # type: ignore [no-any-return]


is_instance = is_attrs_class


def asdict(obj: AttrsInstance) -> dict[str, Any]:
    import attrs

    return attrs.asdict(obj)


def astuple(obj: AttrsInstance) -> tuple[Any, ...]:
    import attrs

    return attrs.astuple(obj)


def replace(obj: AttrsInstance, /, **changes: Any) -> AttrsInstance:
    """Return a copy of obj with the specified changes."""
    import attrs

    return attrs.evolve(obj, **changes)  # type: ignore [misc]


def fields(class_or_instance: Any | type) -> tuple[Field, ...]:
    import attrs

    cls = (
        class_or_instance
        if isinstance(class_or_instance, type)
        else type(class_or_instance)
    )
    fields: list[Field] = []
    for f in attrs.fields(cls):
        f = cast(attrs.Attribute, f)
        default = Field.MISSING if f.default is attrs.NOTHING else f.default
        default_factory = Field.MISSING
        if isinstance(default, attrs.Factory):  # type: ignore [arg-type]
            default_factory, default = default.factory, Field.MISSING
        fields.append(
            Field(
                name=f.name,
                type=f.type,
                default=default,
                default_factory=default_factory,
                repr=f.repr,
                init=f.init,
                compare=f.eq,
                kw_only=f.kw_only,
                hash=f.hash,
                native_field=f,
                metadata=f.metadata,
            )
        )

    return tuple(fields)


def params(obj: AttrsInstance) -> DataclassParams:
    """Return parameters used to define the dataclass."""
    cls = obj if isinstance(obj, type) else type(obj)

    return DataclassParams(
        frozen=getattr(cls.__setattr__, "__name__", None) == "_frozen_setattrs"
    )
